"""
This module contains different functions to identify the DOI (or arXiv identifier) of a paper starting from a PDF file.
The module is divided in two parts. The first part contains low-level functions. These are functions that perform several small tasks 
(e.g., find a DOI in a string of text, validating a DOI, etc.) and that are not supposed to be called directly by the user or by 
any part of the main script main.py. Instead, they are called by the high-level finder functions, defined in the second part of 
this module.
"""
from urllib.parse import unquote
from itertools import accumulate
from PyPDF2 import PdfFileReader, PdfFileWriter
from pdfminer.high_level import extract_text

import requests
import pdftitle
import re
import logging
from googlesearch import search
import pdf2doi.config as config
from pdf2doi import reader_libraries
from pdf2doi.find_title_via_pymupdf import find_title_via_pymupdf
import os
import feedparser

from pdf2doi.patterns import (
    arxiv2007_pattern,
    doi_regexp,
    arxiv_regexp,
    standardise_doi#,
    #isbn_regexp
)

logger = logging.getLogger('pdf2doi')

######## Beginning first part, low-level functions ######## 

def validate_doi_web(doi,method=None):
    """ It queries dx.doi.org for a certain doi, to check that the doi exists.
    If dx.doi.org could not find any paper associated to this doi, the function returns None
    If it was not possible to connect to dx.doi.org, the function returns -1
    If dx.doi.org confirmed that DOI exists, the function returns the full text obtained from dx.doi.org
    Depending on the value of method (default =config.get('method_dxdoiorg')), the format of the text returned by dx.doi.org will be different
    """
    if method == None:
        method = config.get('method_dxdoiorg')
    try:
        # TODO(DJRHails): This should really use the handle API (https://www.doi.org/factsheets/DOIProxy.html)
        url = "http://dx.doi.org/" + doi
        headers = {"accept": method}
        NumberAttempts = 10
        while NumberAttempts:
            r = requests.get(url, headers = headers)
            r.encoding = 'utf-8' #This forces to encode the obtained text with utf-8
            text = r.text
            # 503 or 504 errors are common
            if r.status_code >= 500 or (text.lower().find("503 Service Unavailable".lower() )>=0) or (not text):
                NumberAttempts = NumberAttempts -1
                logger.info("Could not reach dx.doi.org. Trying again. Attempts left: " + str(NumberAttempts))
                continue
            else:
                NumberAttempts = 0

            # 404 = DOI Not Found, or DOI Prefix Not Found
            if r.status_code == 404:
                return None

            # Backup check for HTML error page content
            if text.lower().find("DOI cannot be found".lower()) != -1:
                return None
                
            return text
    except Exception as e:
        logger.error(r"Some error occured within the function validate_doi_web")
        logger.error(e)
        return -1

def validate_arxivID_web(arxivID):
    """It queries export.arxiv.org for a certain arxiv ID, to check that it exists.
    If export.arxiv.org could not find any paper associated to this arxiv ID, the function returns None
    If it was not possible to connect to export.arxiv.org, the function returns -1
    If export.arxiv.org confirmed that DOI exists, the function returns the data obtained from export.arxiv.org
    """
    try:
        url = "http://export.arxiv.org/api/query?search_query=id:" + arxivID
        result = feedparser.parse(url)
        items = result.entries[0]
        found = len(items) > 0
        if not found: 
            return None
        else:
            return items
    except Exception as e:
        logger.error(r"Some error occured within the function arxiv2bib")
        logger.error(e)
        return -1    

def validate(identifier,what='doi'):
    """
    Check that an identifier is a valid identifier by using a regular expression.
    If config.get('webvalidation') == true, it also checks that the identifier is actually associate to a paper via 
    a query to the proper website (e.g. http://dx.doi.org/ for DOIs).

    Parameters
    ----------
    identifier : string
        The identifier (e.g. doi or arXiv ID) to validate.
    what : string, optional. Possible values = 'doi' (default), 'arxiv'
        Specifies the kind of identifier to be validated.
    Returns
    -------
    result :    If config.get('webvalidation') is set to true but connection was not possible, it returns None
                If config.get('webvalidation') is set to true and the identifier is validated, it returns the validation object (could be either a string or dictionary)
                If config.get('webvalidation') is set to false, it returns a single boolean value (True if identifier is valid, False if is not valid)
    """  
    if not identifier:
        return None
    if what=='doi':
        standard_doi = standardise_doi(identifier)
        if identifier != standard_doi:
            logger.info(f"Standardised DOI: {identifier} -> {standard_doi}")

        if standard_doi:
            if config.get('webvalidation'):
                logger.info(f"Validating the possible DOI {standard_doi} via a query to dx.doi.org...")
                result = validate_doi_web(standard_doi)
                if result==-1:
                    logger.error(f"Some error occured during connection to dx.doi.org.")
                    return None
                if isinstance(result,str) and result.strip()[0:5] == '@misc':
                    logger.error(f"The DOI was validated by by dx.doi.org, but the validation string starts with the tag \"@misc\". This might be the DOI of the journal and not the article itself.")
                    return False
                if result:
                    logger.info(f"The DOI {standard_doi} is validated by dx.doi.org.")
                    return result
                else:
                    logger.info(f"The DOI {standard_doi} is not valid according to dx.doi.org.")
                    return False
            else:
                logger.info(f"NOTE: Web validation is deactivated. Set webvalidation = True (or remove the '-nwv' argument if working from command line) in order to validate a potential DOI on dx.doi.org.")
                return True
        else: return False

    elif what=='arxiv':
        if re.match(arxiv2007_pattern,identifier,re.I):
            if config.get('webvalidation'):
                logger.info(f"Validating the possible arxiv ID {identifier} via a query to export.arxiv.org...")
                result = validate_arxivID_web(identifier)
                if result==-1:
                    logger.error(f"Some error occured during connection to export.arxiv.org.")
                    return None
                if result:
                    logger.info(f"The arXiv ID {identifier} is validated by export.arxiv.org")
                    if 'arxiv_doi' in result.keys():
                        logger.info(f"Moreover, export.arxiv.org told us that this paper actually has a DOI: {result['arxiv_doi']}")
                    return result
                else:
                    logger.info(f"The arXiv ID {identifier} is not valid according to export.arxiv.org.")
                    return False
            else:
                logger.info(f"NOTE: Web validation is deactivated. Set webvalidation = True (or remove the '-nwv' argument if working from command line) in order to validate a potential arxiv ID on export.arxiv.org.")
                return True
        else: return False

    #elif what=='isbn':
    #    # code taken from https://www.oreilly.com/library/view/regular-expressions-cookbook/9781449327453/ch04s13.html
    #    # Remove non ISBN digits, then split into a list
    #    chars = list(re.sub("[- ]|^ISBN(?:-1[03])?:?", "", identifier))
    #    # Remove the final ISBN digit from `chars`, and assign it to `last`
    #    last = chars.pop()
    #    if len(chars) == 9:
    #        # Compute the ISBN-10 check digit
    #        val = sum((x + 2) * int(y) for x,y in enumerate(reversed(chars)))
    #        check = 11 - (val % 11)
    #        if check == 10:
    #            check = "X"
    #        elif check == 11:
    #            check = "0"
    #    else:
    #        # Compute the ISBN-13 check digit
    #        val = sum((x % 2 * 2 + 1) * int(y) for x,y in enumerate(chars))
    #        check = 10 - (val % 10)
    #        if check == 10:
    #            check = "0"
    #    if (str(check) == last):
    #        return True
    return False


def extract_arxivID_from_text(text,version=0):   
    """
    It looks for an arxiv ID in the input argument 'text', by using the regexp specified by arxiv_regexp[version],
    where arxiv_regexp is a list of strings defined in patterns.py and 'version' is an integer value.

    Parameters
    ----------
    text : string
        Text to analyse
    version : integer, optional
        Numerical value defined between 0 and len(arxiv_regexp)-1. It specifies which element of the list
        arxiv_regexp is used for the regular expression

    Returns
    -------
    arxivID : list
        It returns a list of all arXiv IDs found (or empty list if no ID was found)
    """   
    try:                                            
        arxiv_ids = re.findall(arxiv_regexp[version] ,text,re.I)
        return arxiv_ids
    except:
        pass
    return []

def extract_doi_from_text(text,version=0):
    """
    It looks for a DOI in the input argument 'text', by using the regexp specified by doi_regexp[version],
    where doi_regexp is a list of strings defined in patterns.py and 'version' is an integer value.

    Parameters
    ----------
    text : string
        Text to analyse
    version : integer, optional
        Numerical value defined between 0 and len(doi_regexp)-1. It specifies which element of the list
        doi_regexp is used for the regular expression

    Returns
    -------
    doi_list : list
        It returns a list of all dois found (or empty list if no doi was found)

    """    
    try:
        dois = re.findall(doi_regexp[version],text,re.I)
        return dois
    except:
        pass
    return []

#def extract_isbn_from_text(text,version=0):
#    """
#    It looks for a ISBN in the input argument 'text', by using the regexp specified by isbn_regexp[version],
#    where isbn_regexp is a list of strings defined in patterns.py and 'version' is an integer value.

#    Parameters
#    ----------
#    text : string
#        Text to analyse
#    version : integer, optional
#        Numerical value defined between 0 and len(doi_regexp)-1. It specifies which element of the list
#        isbn_regexp is used for the regular expression

#    Returns
#    -------
#    isbn_list : list
#        It returns a list of all ISBNs found (or empty list if no one was found)

#    """    
#    try:
#        isbns = re.findall(isbn_regexp[version],text,re.I)
#        return isbns
#    except:
#        pass
#    return []

def find_identifier_in_google_search(query,func_validate,numb_results):
    headers = {"User-Agent":"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36"}

    MaxLengthDisplay = 100
    if len(query)>MaxLengthDisplay:
        query_to_display = query[0:MaxLengthDisplay]  + " ...[query too long, the remaining part is suppressed in the logging]"
    else:
        query_to_display = query
    logger.info(f"Performing google search with key \"" + query_to_display + "\"")
    logger.info(f"and looking at the first {numb_results} results...")
    i=1
    try:
        for url in search(query, stop=numb_results):
            identifier,desc,info = find_identifier_in_text([url],func_validate)
            if identifier: 
                logger.info(f"A valid {desc} was found in the search URL.")
                return identifier,desc,info
            logger.info(f"Looking for a valid identifier in the search result #{str(i)} : {url}")
            response = requests.get(url,headers=headers)
            text = response.text
            identifier,desc,info = find_identifier_in_text(text,func_validate)
            if identifier: 
                return identifier,desc,info
            i=i+1
    except Exception: 
        logger.exception('Some error occured while doing a google search (maybe the string is too long?)')
    return None, None, None

def find_identifier_in_text(texts,func_validate):
    """
    Given any string (or list of strings), it looks for any pattern which matches a valid identifier (e.g. a DOi or an arXiv ID). 
    If a list of string is passed as an argument, they are checked in the order in which they appear in the list, 
    and the function stops as soon as a valid identifier is found.

    Parameters
    ----------
    texts : string or a list of strings
       text to analyse.
    func_validate : function
        The function func_validate is used to validate any identifier that is found. 
        It must take two input arguments, first one being the identifier to validate and the second one being the type
        of identifier (e.g. 'doi,''arxiv') and return False when the identifier is not valid and either True or a non-empty string
        when the identifier is valid. For most application func_validate can be set equal to the function validate defined in this same module.  
        
    Returns
    -------
    identifier : string
        A valid identifier if any is found, or None if nothing was found.
    desc : string
        Description of what was found (e.g. 'doi,''arxiv')
    validation : string or True
        The output returned by the function func_validate. If web validation is enabled, this is typically a bibtex entry for this 
        publication. Otherwise, it is just equal to True

    """
    if not isinstance(texts,list): texts = [texts]
    
    for text in texts:

        if isinstance(text, bytes):
            text = text.decode()

        #First we look for DOI
        for v in range(len(doi_regexp)):
            identifiers = extract_doi_from_text(text,version=v)
            for identifier in identifiers:
                logger.debug("Found a potential DOI: " + identifier)
                validation = func_validate(identifier,'doi')
                standard_doi = standardise_doi(identifier)
                if identifier != standard_doi:
                    logger.info(f"Standardised DOI: {identifier} -> {standard_doi}")

                if validation: 
                    return standard_doi, 'DOI', validation
            
        #Then we look for an Arxiv ID
        for v in range(len(arxiv_regexp)):
            identifiers = extract_arxivID_from_text(text,version=v)
            for identifier in identifiers:
                validation = func_validate(identifier,'arxiv')
                if validation:
                    return identifier,'arxiv ID', validation

        ##Then we look for ISBNs
        #for v in range(len(isbn_regexp)):
        #    identifiers = extract_isbn_from_text(text,version=v)
        #    for identifier in identifiers:
        #        validation = func_validate(identifier,'isbn')
        #        if validation:
        #            return identifier,'isbn', validation

    return None, None, None


def get_pdf_info(file):
    """
    Given a valid file object, it returns a dictionary of info. 
    Currently, it uses PyPDF to extract the info.
    
    Parameters
    ----------
    file : file object, opened as 'rb
    
    Returns
    -------
    info : dictionary
    """

    try:
        pdf = PdfFileReader(file,strict=False)
    except Exception as e:
        logger.error("It was not possible to open the file with PyPDF2. Is this a valid pdf file?")
        logger.error(f"{e}")
        return None
    try:
        info = pdf.getDocumentInfo()
    except Exception as e:
        logger.error(f"An error occurred when retrieving the pdf info with PyPDF2: {e}")
        return None

    #file.close()
    return info

    
def find_possible_titles(file):
    """
    Given a valid file object, it tries to extract a list of possible titles. 
    In the current implementation it looks for titles by 1) looking for the outcome of pdftitle library, 
    2) looking in the dictionary returned by the PyPDF library and 3) looking in the filename.

    Parameters
    ----------
    file : file object, opened as 'rb
    
    Returns
    -------
    titles : list of strings
        Possible titles of the paper.
    """
    titles = []
    # (1)    
    try:
        title = pdftitle.get_title_from_io(file)
    except:
        title = ''
    if isinstance(title, str):
        if len(title.strip())>12:#This is to check that the title found is neither empty nor just few characters
            logger.info(f"pdftitle found the title \"{title}\"")
            titles.append(title.strip())  
    # (2)    
    try:
        title = find_title_via_pymupdf(file)
    except:
        title = ''
    if isinstance(title, str):
        if len(title.strip())>12:#This is to check that the title found is neither empty nor just few characters
            logger.info(f"pymupdf found the title \"{title}\"")
            titles.append(title.strip())  
        
    # (3)
    info = get_pdf_info(file)
    if info:
        for key, value in info.items():
            if 'title' in key.lower():
                if isinstance(value,str) and len(value.strip())>12 and len(value.split())>3: #This is to check that the title found is neither empty nor just few characters or few words
                    logger.info(f"PyPDF2 found the title \"{title}\"")
                    titles.append(value.strip())         
    # (4)
    title = os.path.basename(file.name)
    if len(title.strip())>30:#This is to check that the title found is neither empty nor just few characters
        logger.info(f"Using the filename as potential title, \"{title}\"")
        titles.append(title.strip())
        
    # remove possible duplicates
    titles_temp = titles
    titles = []
    [titles.append(x) for x in titles_temp if x not in titles]
    return titles
        
def get_pdf_text(file,reader):
    """
    Given a valid file object (pointing to a pdf file), it returns the text of the pdf file extracted with the library 
    specified in the 'reader' input variable.

    Parameters
    ----------
    file : file object, opened as 'rb
    reader : string
        It specifies which library is used to extract the text from the pdf.
        The currently supported values are either 
            'pypdf' (uses the PyPDF2 module)
            'textract' (uses the 'textract' module)
    Returns
    -------
    text : list of strings
    """
    text =[]

    if reader == 'pdfminer':
        try:
            pdf_text = extract_text(file)
        except Exception as e:
            logger.error(f"An error occurred when reading the content of this file with pdfminer.")
            logger.error("Error from pdfminer: " + str(e))
            return None
            
        text.append(pdf_text)

    if reader == 'pypdf':
        try:
            pdf = PdfFileReader(file,strict=False)
        except Exception as e:
            logger.error(f"An error occurred when reading the content of this file with PyPDF2.")
            logger.error("Error from PyPDF2: " + str(e))
            return None
        try:
            number_of_pages = pdf.getNumPages()
        except Exception as e:
            logger.error(f"An error occurred when retrieving the number of pages in the pdf with PyPDF2.")
            logger.error("Error from PyPDF2: " + str(e))
            return None
            
        for i in range(number_of_pages):
            try:
                text.append( (pdf.getPage(i)).extract_text())
            except Exception as e:
                logger.error("An error occured while loading the document text with PyPDF2. The pdf version might be not supported.")
                logger.error("Error from PyPDF2: " + str(e))
                break 

        # Checking if there are annotations
        for page in pdf.pages:
            if "/Annots" in page:
                for annot in page["/Annots"]:
                    subtype = annot.get_object()["/Subtype"]
                    if subtype in ["/FreeText", "/Text"]:
                        text.append(annot.get_object()["/Contents"])

    if reader == 'textract':
        import textract
        #This block of code is not very efficient. We start from an object file (file), we extract its path, and then we pass this path to the library
        #textract, which will later re-open the file; however, right now there isn't a workaround, because textract does not accept an object file as input.
        path = file.name #Note: this part will fail with if the object file does not correpond to a locally available file
        try:
            text = [textract.process(path,encoding='utf-8', errors='ignore').decode('utf-8')]
        except Exception as e:
            logger.error(e)
            try:
                text = [textract.process(path)]
            except Exception as e:   
                logger.error("An error occured while loading the document text with textract. The pdf version might be not supported.")
                logger.error("Error from textract: " + str(e))
    return text

def add_metadata(target,key,value):
    """Given a pdf file or a folder identified by the input variable target, it adds a metadata with label equal to key
    and containing the content of the input variable value to all pdf files specified by target (either a single file or
    all the pdf files in a folder). 

    Parameters
    ----------
    target : string
        a valid path to a pdf file or a folder
    key : string
    value : string
    Returns
    -------
    True if the the metadata was added succesfully, false otherwise
    """
    list_files = []
    if  os.path.isdir(target): #if target is a folder, we populate the list list_files with all the pdf files contained in this folder
        logger.info(f"Looking for pdf files in the folder {target}...")
        pdf_files = [f for f in os.listdir(target) if (f.lower()).endswith('.pdf')]
        numb_files = len(pdf_files)
        if len(pdf_files) == 0:
            logger.error("No pdf files found in this folder.")
            return None
        logger.info(f"Found {numb_files} pdf files.")
        if not(target.endswith(config.get('separator'))): #Make sure the path ends with "\" or "/" (according to the OS)
            target = target + config.get('separator')
        for f in pdf_files:
            list_files.append(target + f)
    else:
        list_files = [target]

    for f in list_files:
        logger.info(f"Trying to add the tag \'{key}\'-> \'{value}\' into the metadata of the file \'{f}\'...")
        try:
            file = open(f, 'rb') 
        except (FileNotFoundError, IOError):
            msg = "File not found."
            logger.error(msg)
            return False, msg
        try:
            pdf = PdfFileReader(f,strict=False)
        except:
            msg = "It was not possible to open the file with PyPDF2. Is this a valid pdf file?"
            logger.error(msg)
            return False, msg
        try:
            writer = PdfFileWriter()
            writer.appendPagesFromReader(pdf)
            metadata = pdf.getDocumentInfo()
            try:
                writer.add_metadata(metadata)    #This instruction might generate an error if the pre-existing metadata are weird and are not
            except:                             #correctly seen as strings (it happens with old files). Therefore we use the try/except
                pass                            #to ignore this possible problem
            writer.add_metadata({
                key: value
            })
            fout = open(f, 'ab') 
            writer.write(fout)
            file.close()
            fout.close()
            logger.info(f"The tag \'{key}\'-> \'{value}\' was added succesfully to the metadata of the file \'{f}\'...")
        except Exception as e:
            logger.error("Error from PyPDF2: " + str(e))
            msg = f"An error occured while trying to write the tag \'{key}\'-> \'{value}\'  into the metadata of the file \'{f}\'. Maybe the file is open elsewhere?"
            logger.error(msg)
            return False, msg

def add_found_identifier_to_metadata(target,identifier):
    """Given a pdf file or a folder identified by the input variable target, it adds a metadata with label '/pdf2doi_identifier'
    and containing the content of the input variable identifier to all pdf files specified by target (either a single file or
    all the pdf files in a folder). This can be useful to make sure that the next time
    this same pdf is analysed, the identifier is found more easily.
    It can also be useful when one want to reset to '' the '/pdf2doi_identifier' of all pdf files in a certain folder.

    Parameters
    ----------
    target : string
        a valid path to a pdf file or a folder
    identifier : string
        a valid identifier, which will be stored in the pdf metadata with name '/pdf2doi_identifier'
    Returns
    -------
    True if the the metadata was added succesfully, false otherwise
    """
    add_metadata(target,key='/pdf2doi_identifier',value=identifier)


######## End first part ######## 

###########################################################################################################
###########################################################################################################
###########################################################################################################

######## Beginning second part, high-level functions ######## 

'''
The following functions are high-level identifier finders, i.e. functions that can be called directly by the user or by the function pdf2doi
contained in main.py. The function find_identifier acts as a wrapper for all the other high-level functions: 
based on the value of the input argument method, it call the corresponding function. 
The correspondence between the values of the string method and the function to call is defined in the dictionary finder_methods 
(see bottom of this module).
'''

def find_identifier(file, method, func_validate=validate,**kwargs):
    """ Tries to find an identifier (e.g. DOI, arxiv ID,...) for the pdf file identified by the input
    argument 'file', by using the method specified by input argument 'method'. Any found identifier is validated
    by using the function func_validate. If a valid identifier is found with any method different from
    "document_infos" (i.e. by looking into the file metadata) the identifier is also added to the file metadata
    with key "/pdf2doi_identifier" (unless config.get('save_identifier_metadata') is set to False)

    Parameters
    ----------
    file : object file 
    method : string
        Method to be used to look for an identifier. The possible values are specified by the keys of 
        the dictionary finder_methods.
    func_validate : function, optional, the default value is the validate() function defined in this module
        The function func_validate is used to validate any identifier that is found. 
        It must take two input arguments, first one being the identifier to validate and the second one being the type
        of identifier (e.g. 'doi,''arxiv') and return False when the identifier is not valid and either True or a non-empty string
        when the identifier is valid. 
    kwargs : 
        additional input parameters specific to the method chosen
        
    Returns
    -------
    result : dictionary
        result['identifier'] = DOI or other identifier (or None if nothing is found)
        result['identifier_type'] = string specifying the type of identifier (e.g. 'doi' or 'arxiv')
        result['validation_info'] = Additional info on the paper. If config.get('webvalidation') = True, then result['validation_info']
                                    will typically contain a raw bibtex data for this paper. Otherwise it will just contain True                         
        result['path'] = path of the pdf file
        result['method'] = method used to find the identifier

    """
    if not method in  finder_methods.keys():
        raise ValueError("The input variable method is not valid. Possible values = \'" + "\',\'".join(finder_methods.keys())+"\'")
    if not callable(func_validate):  func_validate = lambda x : x

    identifier, desc, info = finder_methods[method](file,func_validate,**kwargs)
    
    ### The next block of code check if the identifier found is an arXiv ID, and tried to replace it with a DOI (either from a journal publication or with the arXiv DOI)
    ### If the new DOI is from a journal, it also validates it again, in order to store the correct bibtex info inside the result['info'] string 
    ### This needs to be implemented more elegantly, probably in a dedicated function
    if identifier:
        if desc == 'arxiv ID' and config.get('replace_arxivID_by_DOI_when_available') == True:
            if 'arxiv_doi' in info.keys() and info['arxiv_doi']:
                logger.info(f"Checking if the DOI {info['arxiv_doi']} is valid...")
                arxiv_doi = info['arxiv_doi']
                info = validate(arxiv_doi,'doi')
                if info:
                    logger.info(f"The arXiv ID will be replaced by the Journal DOI {identifier}. If you prefer to keep the arXiv ID, use the command -no_arxiv2doi when invoking pdf2doi")
                    desc = 'DOI'
                    identifier = arxiv_doi
                    method = method + ' + arxiv2doi'
            else:
                arxiv_doi = f"10.48550/arXiv.{identifier}"
                logger.info(f"The arXiv ID will be replaced by the arXiv DOI {arxiv_doi}. If you prefer to keep the arXiv ID, use the command -no_arxiv2doi when invoking pdf2doi")
                desc = 'arxiv DOI'
                identifier = arxiv_doi
                method = method + ' + arxiv2doi'
            
    ######     
            
    result = {'identifier':identifier,'identifier_type':desc,
              'path':file.name, 'method':method}
    result['validation_info'] = info

    return result

def find_identifier_in_pdf_info(file,func_validate,keysToCheckFirst=[]):
    """ 
    Try to find a valid DOI in the values of the 'document information' dictionary. If a list of string is specified via the optional
    input parameter 'keysToCheckFirst', then the corresponding elements of the dictionary (assuming that the key exists) are given
    priority.

    Parameters
    ----------
    file : object file 
    keysToCheckFirst : list, optional
        A list of strings. If specified, element of the 'document information' dictionary with these keys (assuming that 
       they exist) are given priority in the DOI search (following the order they appear in this list)
    Returns
    -------
    result : dictionary with identifier and other info (see above) 
    """


    #This is a list of keys that will NOT be considered when looking for an identifier in the metadata. 
    #Some of them are known to contain doi-like patterns but not the actual doi of the publication.
    #For example '/wps-journaldoi' contains the DOI of the journal
    KeysNotToUse = ['/wps-journaldoi'] 
    pdfinfo = get_pdf_info(file)
    identifier, desc, info = None, None, None
    if pdfinfo:
        Keys = keysToCheckFirst + list(pdfinfo.keys()) #Quick way to create a list of keys where the elements of keysToCheckFirst appear first. 
                                            #Some elements of keysToCheckFirst might be duplicated in 'Keys', but this is not a problem
                                            #because the corresponding element in the pdfinfo dictionary is eliminated after being checked
        for key in Keys:
            if key in pdfinfo.keys() and key.lower() not in KeysNotToUse:
                identifier,desc,info = find_identifier_in_text(pdfinfo[key],func_validate)
                if identifier: 
                    logger.info(f"A valid {desc} was found in the document info labelled \'{key}\'.")
                    break
                del pdfinfo[key]
    if identifier:
        return identifier,desc,info
    else:
        logger.info("Could not find a valid identifier in the document info.")
        return None, None, None

def find_identifier_in_filename(file, func_validate):
    """ 
    Parameters
    ----------
    file : object file 
    func_validate : function
    Returns
    -------
    result : dictionary with identifier and other info (see above)
    """
    # We unquote here to allow / to be placed in filenames as %2F
    # particularly useful for DOIs with multiple slashes
    text = unquote(os.path.basename(file.name))

    # 10.1227/12345678.pdf is both a valid filepath and a valid doi
    # We want to still discover the "actual" doi which requires the .pdf extension to be removed.
    # We try repeatedly stripping the extension until we find a valid doi (works best with online validation, which is default). 
    strip_possible_extensions = list(accumulate(text.split('.'), lambda x,y: '.'.join([x, y])))
    texts = list(reversed(strip_possible_extensions))

    identifier,desc,info = find_identifier_in_text(texts,func_validate)
    if identifier: 
        logger.info(f"A valid {desc} was found in the file name.")
        return identifier,desc,info
    else:
        logger.info("Could not find a valid identifier in the file name.")
        return None, None, None

def find_identifier_in_pdf_text(file, func_validate):
    """ Try to find a valid identifier in the plain text of the pdf file. The text is extracted via the function
    get_pdf_text.

    Parameters
    ----------
    file : object file 
    func_validate : function
    Returns
    -------
    result : dictionary with identifier and other info (see above)
    """
    for reader in reader_libraries:
        logger.info(f"Extracting text with the library {reader}...")
        texts = get_pdf_text(file,reader.lower())
        
        if not isinstance(texts,list):
            texts = [texts]
        if texts:
            logger.info(f"Text extracted succesfully. Looking for an identifier in the text...")
            identifier,desc,info = find_identifier_in_text(texts,func_validate)
            if identifier: 
                logger.info(f"A valid {desc} was found in the document text.")
                return identifier,desc,info
            else:
                logger.info(f"Could not find a valid identifier in the document text extracted by {reader}.")
    logger.info("Could not find a valid identifier in the document text.")
    return None, None, None

def find_identifier_by_googling_title(file, func_validate):
    """
    Parameters
    ----------
    file : object file 
    func_validate : function, optional
    """
    logger.info(f"Trying to find the title of this publication...")
    titles = find_possible_titles(file)
    if titles:
        if config.get('websearch')==False:
            logger.info("NOTE: Possible titles of the paper were found, but the web-search method is currently disabled by the user. Enable it in order to perform a qoogle query.")
            return None, None, None
        else:
            logger.info(f"Found {len(titles)} possible title(s).")
            titles.sort(key=len, reverse=True)
            for index_title, title in enumerate(titles):
                logger.info(f"Trying possible title #{index_title+1} '{title}'")
                identifier,desc,info = find_identifier_in_google_search(title,func_validate,numb_results=config.get('numb_results_google_search'))
                if identifier:
                    logger.info(f"A valid {desc} was found with this google search.")
                    return identifier, desc, info
            logger.info("None of the search results contained a valid identifier.")     
            return None, None, None
    else:
        logger.error("It was not possible to find a title for this file.")
        return None, None, None
    
def find_identifier_by_googling_first_N_characters_in_pdf(file, func_validate, numb_results=config.get('numb_results_google_search'), numb_characters=config.get('N_characters_in_pdf')):
    """
    Parameters
    ----------
    file : object file 
    func_validate : function, optional
    numb_results : int
    numb_characters : int
    """
    if config.get('websearch')==False:
        logger.info("NOTE: Web-search methods are currently disabled by the user. Enable it in order to use this method.")
        return None, None, None

    for reader in reader_libraries:
        logger.info(f"Trying to extract the first {numb_characters} characters from the pdf file by using the library {reader}...")
        text = get_pdf_text(file,reader.lower())
        if text==[] or text=="":
            logger.error(f"The library {reader} could not extract any text from this file.")
            continue
        if isinstance(text,list):
            text = "".join(text)
        if not(isinstance(text, str)):
            logger.error(f"The library {reader} could not extract any text from this file.")
            continue 
        text = re.sub(r'[^\x00-\x7f]',r' ',text)    #Remove all non-text characters from the string text
        for r in ("\n","\r","\t"):
            text = text.replace(r," ")
         
        if text=="":                                #Check tha the string is still not empty after removing non-text characters
            logger.error(f"The library {reader} could not extract any meaningful text from this file.")
            continue
            
        text = text[0:numb_characters]              #Select the first numb_characters characters

        logger.info(f"Doing a google search, looking at the first {config.get('numb_results_google_search')} results...")
        identifier,desc,info = find_identifier_in_google_search(text,func_validate,numb_results)
        if identifier:
            logger.info(f"A valid {desc} was found with this google search.")
            return identifier,desc,info

    logger.info(f"Could not find a valid identifier by googling the first {numb_characters} characters extracted from the pdf file.")
    return None, None, None
        
#The dictionary finder_methods list all the methods currently implemented to find an identifier.
#Each method is associated to a function.
#The keys of this dictionary corresponds to the possible values of the input argument 'method' 
#passed the function find_identifier
finder_methods = {
                    "document_infos"            : find_identifier_in_pdf_info,
                    "document_text"             : find_identifier_in_pdf_text,
                    "filename"                  : find_identifier_in_filename,
                    "title_google"              : find_identifier_by_googling_title,
                    "first_N_characters_google" : find_identifier_by_googling_first_N_characters_in_pdf
                    }

######## End second part ######## 



# This regex for the DOI is taken from https://www.crossref.org/blog/dois-and-matching-regular-expressions/
# but allows multiple separator types, a prefix, and assumes the DOI is lowercase.
import re

# Based on local DOI corpus:
# 0% have non-standard separators (e.g. 10.1177:0146167297234003)
# 0.0009% have a non-standard namespace (i.e. not \d{4,9}), all match [\d\.]{1,15}
# 32% have interesting characters (i.e. not [.\/a-z0-9])
# 99.34% match this pattern: '^10\.\d{4,9}/[-._;()\/:a-z0-9][a-z0-9]$'
# 99.43% match this pattern: '^10\.\d{4,9}/[-._;()\/:a-z0-9]+$', crossref claim 99.3%
# 99.880% terminate in alphanumeric characters [a-z0-9]
# 99.804% match: ^10[.][\d.]{1,15}/[-._;:()/a-z0-9<>]+[a-z0-9]$
# 99.9999% match: ^10[.][\d.]{1,15}/[-._;()/:a-z0-9<>@\]\[#+?\s&]+$
DOI = r"""(?xm)
  (?P<marker>   doi[:\/\s]{0,3})?
  (?P<prefix>
    (?P<namespace> 10)
    [.]
    (?P<registrant> \d{2,9})
  )
  (?P<sep>     [:\-\/\s\]])
  (?P<suffix>  [\-._;()\/:a-z0-9]+[a-z0-9]) 
  (?P<trailing> ([\s\n\"<.]|$))
"""

def standardise_doi(identifier):
    """
    Standardise a DOI by removing any marker, lowercase, and applying a consistent separator
    """
    doi_meta = dict()
    for match in re.finditer(DOI, identifier.lower()):
        doi_meta.update(match.groupdict())
    
    if any(key not in doi_meta for key in ["registrant", "suffix"]):
        return None
    
    return f"10.{doi_meta['registrant']}/{doi_meta['suffix']}"


# This is a regex for arxiv identifiers (in use after 2007) and it is used to validate a given arxiv ID.
arxiv2007_pattern = '^(\d{4}\.\d+)(?:v\d+)?$'
                                                                            

# The list doi_regexp contains several regular expressions used to identify a DOI in a string. They are (roughly) ordered from stricter to less and less strict.
doi_regexp = ['doi[\s\.\:]{0,2}(10\.\d{4}[\d\:\.\-\/a-z]+)(?:[\s\n\"<]|$)', # version 0 looks for something like "DOI : 10.xxxxS[end characters] where xxxx=4 digits, S=combination of characters, digits, ., :, -, and / of any length
                                                                            # [end characters] is either a space, newline, " , < or the end of the string. The initial part could be either "DOI : ", "DOI", "DOI:", "DOI.:", ""DOI:." 
                                                                            # and with possible spaces or lower cases.
              '(10\.\d{4}[\d\:\.\-\/a-z]+)(?:[\s\n\"<]|$)',                 # in version 1 the requirement of having "DOI : " in the beginning is removed
              '(10\.\d{4}[\:\.\-\/a-z]+[\:\.\-\d]+)(?:[\s\na-z\"<]|$)',     # version 2 is useful for cases in which, in plain texts, the DOI is not followed by a space, newline or special characters,
                                                                            #but is instead followed by other letters. In this case we can still isolate the DOI if we assume that the DOI always ends with numbers
              'https?://[ -~]*doi[ -~]*/(10\.\d{4,9}/[-._;()/:a-z0-9]+)(?:[\s\n\"<]|$)', # version 3 is useful when the DOI can be found in a google result as an URL of the form https://doi.org/[DOI]
                                                                            #The regex for [DOI] is 10\.\d{4,9}/[-._;()/:a-z0-9]+ (taken from here https://www.crossref.org/blog/dois-and-matching-regular-expressions/)
                                                                            #and it must be followed by a valid ending character: either a speace, a new line, a ", a <, or end of string.
              '^(10\.\d{4,9}/[-._;()/:a-z0-9]+)$']                         # version 4 is like version 3, but without the requirement of the url https://doi.org/ in front of it.
                                                                            #However, it requires that the string contains ONLY the doi and nothing else. This is useful for when the DOI is stored in metadata

             
#Similarly, arxiv_regexp is a list of regular expressions used to identify an arXiv identifier in a string. They are (roughly) ordered from stricter to less and less strict. Moreover,
#the regexp corresponding to older arXiv notations have less priority.
#NOTE: currently only one regexp is implemented for arxiv. 
arxiv_regexp = ['arxiv[\s]*\:[\s]*(\d{4}\.\d+)(?:v\d+)?(?:[\s\n\"<]|$)',  #version 0 looks for something like "arXiv:YYMM.number(vn)" 
                                                                            #where YYMM are 4 digits, numbers are up to 5 digits and (vn) is
                                                                            #an additional optional term specifying the version. n is an integer starting from 1
                                                                            #This is the official format for Arxiv indentifier starting from 1 April 2007
                                                                            #This is the official format for Arxiv indentifier starting from 1 April 2007
                '(\d{4}\.\d+)(?:v\d+)?(?:\.pdf)',                         #version 1 is similar to version 0, without the requirement of "arxiv : " at the beginning 
                                                                            #but with the requirement that '.pdf' appears right after the possible arXiv identifier. 
                                                                            #This is helpful when we are trying to extrat the arXiv ID from the file name.
                '^(\d{4}\.\d+)(?:v\d+)?$']                               #version 2 is similar to version 0, without the requirement of "arxiv : " at the beginning 
                                                                            #but requires that the string contains ONLY the arXiv ID.


##Following regexp is taken from https://www.oreilly.com/library/view/regular-expressions-cookbook/9781449327453/ch04s13.html
#isbn_regexp = [''.join(['(?:ISBN(?:-1[03])?:? )?(?=[0-9X]{10}$|' ,
#                        '(?=(?:[0-9]+[- ]){3})[- 0-9X]{13}$|97[89][0-9]{10}$|' , 
#                        '(?=(?:[0-9]+[- ]){4})[- 0-9]{17}$)(?:97[89][- ]?)?[0-9]{1,5}[- ]?' ,
#                        '[0-9]+[- ]?[0-9]+[- ]?[0-9X]$']),
#               ''.join(['\bISBN(?:-1[03])?:? (?=[0-9X]{10}$|(?=(?:[0-9]+[- ]){3})',
#                        '[- 0-9X]{13}$|97[89][0-9]{10}$|(?=(?:[0-9]+[- ]){4})[- 0-9]{17}$)',
#                        '(?:97[89][- ]?)?[0-9]{1,5}[- ]?[0-9]+[- ]?[0-9]+[- ]?[0-9X]\b'])
#               ]


# This code has been adapted from here https://github.com/LouisdeBruijn/Medium/blob/master/PDF%20retrieval/pdf_retrieval.py
# The Entry Point is the function find_title_via_pymupdf (defined at the bottom)
# Try to find a title of a pdf file by parsing the file and looking at html-like elements that start with <h1>

from operator import itemgetter
import fitz
    
def fonts(doc, granularity=False):
    styles = {}
    font_counts = {}
    for page in doc:
        blocks = page.get_text('dict')['blocks']
        for b in blocks:  # iterate through the text blocks
            if b['type'] == 0:  # block contains text
                for l in b["lines"]:  # iterate through the text lines
                    for s in l["spans"]:  # iterate through the text spans
                        if granularity:
                            identifier = "{0}_{1}_{2}_{3}".format(s['size'], s['flags'], s['font'], s['color'])
                            styles[identifier] = {'size': s['size'], 'flags': s['flags'], 'font': s['font'],
                                                    'color': s['color']}
                        else:
                            identifier = "{0}".format(s['size'])
                            styles[identifier] = {'size': s['size'], 'font': s['font']}
    
                        font_counts[identifier] = font_counts.get(identifier, 0) + 1  # count the fonts usage
    
    font_counts = sorted(font_counts.items(), key=itemgetter(1), reverse=True)
    
    if len(font_counts) < 1:
        raise ValueError("Zero discriminating fonts found!")
    
    return font_counts, styles
    
    
def font_tags(font_counts, styles):
    """Returns dictionary with font sizes as keys and tags as value.
    :param font_counts: (font_size, count) for all fonts occuring in document
    :type font_counts: list
    :param styles: all styles found in the document
    :type styles: dict
    :rtype: dict
    :return: all element tags based on font-sizes
    """
    p_style = styles[font_counts[0][0]]  # get style for most used font by count (paragraph)
    p_size = p_style['size']  # get the paragraph's size
    
    # sorting the font sizes high to low, so that we can append the right integer to each tag
    font_sizes = []
    for (font_size, count) in font_counts:
        font_sizes.append(float(font_size))
    font_sizes.sort(reverse=True)
    
    # aggregating the tags for each font size
    idx = 0
    size_tag = {}
    for size in font_sizes:
        idx += 1
        if size == p_size:
            idx = 0
            size_tag[size] = '<p>'
        if size > p_size:
            size_tag[size] = '<h{0}>'.format(idx)
        elif size < p_size:
            size_tag[size] = '<s{0}>'.format(idx)
    
    return size_tag
    
def headers_para(doc, size_tag):
    """Scrapes headers & paragraphs from PDF and return texts with element tags.
    :param doc: PDF document to iterate through
    :type doc: <class 'fitz.fitz.Document'>
    :param size_tag: textual element tags for each size
    :type size_tag: dict
    :rtype: list
    :return: texts with pre-prended element tags
    """
    header_para = []  # list with headers and paragraphs
    first = True  # boolean operator for first header
    previous_s = {}  # previous span
    
    for page in doc:
        blocks = page.get_text("dict")["blocks"]
        for b in blocks:  # iterate through the text blocks
            if b['type'] == 0:  # this block contains text
    
                # REMEMBER: multiple fonts and sizes are possible IN one block
    
                block_string = ""  # text found in block
                for l in b["lines"]:  # iterate through the text lines
                    for s in l["spans"]:  # iterate through the text spans
                        if s['text'].strip():  # removing whitespaces:
                            if first:
                                previous_s = s
                                first = False
                                block_string = size_tag[s['size']] + s['text']
                            else:
                                if s['size'] == previous_s['size']:
    
                                    if block_string and all((c == "|") for c in block_string):
                                        # block_string only contains pipes
                                        block_string = size_tag[s['size']] + s['text']
                                    if block_string == "":
                                        # new block has started, so append size tag
                                        block_string = size_tag[s['size']] + s['text']
                                    else:  # in the same block, so concatenate strings
                                        block_string += " " + s['text']
    
                                else:
                                    header_para.append(block_string)
                                    block_string = size_tag[s['size']] + s['text']
    
                                previous_s = s
    
                    # new block started, indicating with a pipe
                    block_string += "|"
    
                header_para.append(block_string)
    
    return header_para

def find_title_via_pymupdf(file):
    doc = fitz.open(file)
    font_counts, styles = fonts(doc, granularity=False)
    size_tag = font_tags(font_counts, styles)
    elements = headers_para(doc, size_tag)
    for e in elements:
        if e.startswith('<h1>'):
            return (e.lstrip("<h1>")).replace("|", "")


import configparser
import os
import logging

''' 
method_dxdoiorg                         It sets which method is used when querying dx.doi.org to retrieve the bibtex info
                                        Two possible values are 'text/bibliography; style=bibtex' , 'application/x-bibtex' and
                                        'application/citeproc+json'

                                        The 'application/x-bibtex' method was the one originally used in the first version of
                                        pdf2doi. However, since October 2021 this method does not return the 
                                        Journal info, as a result of a bug in dx.doi.org (Last time checked = 2021 Nov 06)
                                        I added the method 'text/bibliography; style=bibtex' to overcome this problem
                                        However, with 'text/bibliography; style=bibtex' the authors string is returned in
                                        the format "LastName1, FirstName1 SecondName1.. and LastName2, FirstName2 SecondName2.. and etc."
                                        which is not the format expect by the script pdf-renamer, which uses pdf2doi 
                                               
                                        pdf2doi automatically parses the string obtained by dx.doi.org differently based on
                                        the value of method_dxdoiorg
                                        The method "application/citeproc+json" is the best one, because it returns everythin as 
                                        a json dictionary, and it requires no parsing
check_online_to_validate

websearch

numb_results_google_search              How many results should it look into when doing a google search

N_characters_in_pdf

save_identifier_metadata                Sets the default value of the global setting save_identifier_metadata
                                        If set True, when a valid identifier is found with any method different than the metadata lookup the identifier
                                        is also written inside the file metadata with key "/identifier". If set False, this does not happen.
                                 
'replace_arxivID_by_DOI_when_available' 

'''


class config():
    __params={'verbose'   :   True,
            'separator' : os.path.sep,
            'method_dxdoiorg' : 'application/citeproc+json',
            'webvalidation' : True,
            'websearch' : True,
            'numb_results_google_search' : 6,
            'N_characters_in_pdf' : 1000,
            'save_identifier_metadata' : True,
            'replace_arxivID_by_DOI_when_available' : True
            }
    __setters = __params.keys()


    @staticmethod
    def update_params(new_params):
        config.__params.update(new_params)

    @staticmethod
    def get(name):
        return config.__params[name]

    @staticmethod
    def set(name, value):
        if name in config.__setters:
             config.__params[name] = value
        else:
            raise NameError("Name not accepted in set() method")
        #Here we define additional actions to perform when specific parameters are modified
        if name == 'verbose':
            # We change the logger verbosity
            if value: loglevel = logging.INFO
            else: loglevel = logging.CRITICAL
            logger = logging.getLogger("pdf2doi")
            logger.setLevel(level=loglevel)

    @staticmethod
    def ReadParamsINIfile():
        '''
        Reads the parameters stored in the file settings.ini, and stores them in the dict self.params
        If the .ini file does not exist, it creates it with the default values.
        '''
        path_current_directory = os.path.dirname(__file__)
        path_config_file = os.path.join(path_current_directory, 'settings.ini')
        if not(os.path.exists(path_config_file)):
            config.WriteParamsINIfile()
        else:
            config_object = configparser.ConfigParser()
            config_object.optionxform = str
            config_object.read(path_config_file)
            config.__params.update(dict(config_object['DEFAULT']))
            config.ConvertParamsToBool()
            config.ConvertParamsToNumb()

    @staticmethod
    def ConvertParamsToBool():
        for key,val in config.__params.items():
            if isinstance(val, str):
                if val.lower() == 'true':
                    config.__params[key]=True
                if val.lower() == 'false':
                    config.__params[key]=False

    @staticmethod
    def ConvertParamsToNumb():
        for key,val in config.__params.items():
            if isinstance(val, str) and val.isdigit():
                config.__params[key]=int(val)
    @staticmethod
    def print():
        '''
        Prints all settings
        '''
        for key,val in config.__params.items():
            print(key + " : " + str(val) + ' ('+type(val).__name__+')')

    @staticmethod
    def WriteParamsINIfile():
        '''
        Writes the parameters currently stored in in the dict self.params into the file settings.ini
        '''
        path_current_directory = os.path.dirname(__file__)
        path_config_file = os.path.join(path_current_directory, 'settings.ini')
        config_object = configparser.ConfigParser()
        config_object.optionxform = str
        config_object['DEFAULT'] = config.__params
        with open(path_config_file, 'w') as configfile: #Write them on file
            config_object.write(configfile)


###########################################











import logging
import importlib.util

# Setup logging
logger = logging.getLogger("pdf2doi")
logger.setLevel(level=logging.INFO)
if not logger.handlers:
    formatter = logging.Formatter("[pdf2doi]: %(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
logger.propagate = False

from .config import config
config.ReadParamsINIfile()  #Load all current configuration from the .ini file. If the .ini file is not present, it generates it using default values

#Determine the list of libraries to be used to extract text from pdf files
reader_libraries = ['PyPdf','pdfminer'] 
# Using PyPdf before pdfminer makes sure that, in arxiv pdf files, the DOI which is sometimes written on the left margin of the first page is correctly detected

is_textract_installed = importlib.util.find_spec('textract')
if is_textract_installed:
    reader_libraries. append('textract')
    

config.set('verbose',config.get('verbose')) #This is a quick and dirty way (to improve in the future) to make sure that the verbosity of the pdf2doi logger is properly set according
                                            #to the current value of config.get('verbose') (see config.py file for details)
from .main import pdf2doi, pdf2doi_singlefile
from .finders import *
#from .bibtex_makers import *
from .utils_registry import install_right_click, uninstall_right_click





import argparse
import logging
from os import path, listdir
import pdf2doi.finders as finders
import pdf2doi.config as config
import io


# import easygui Modules that are commented here are imported later only when needed, to improve start up time
# import pyperclip

def pdf2doi(target):
    ''' This is the main routine of the library. When the library is used as a command-line tool (via the entry-point "pdf2doi") the input arguments
    are collected, validated and sent to this function (see the function main() below).
    The function tries to extract the DOI (or other identifiers) of the publication in the pdf files whose path is specified in the input variable target.
    If target contains the valid path of a folder, the function tries to extract the DOI/identifer of all pdf files in the folder.
    It returns a dictionary (or a list of dictionaries) containing info(s) about the file(s) examined, or None if an error occurred.

    Example:
        import pdf2doi
        path = r"Path\to\folder"
        result = pdf2doi.pdf2doi(path)
        print(result[0]['identifier'])          # Print doi/identifier of the first pdf file found in this folder
        print(result[0]['identifier_type'])     # Print the type of identifier found (e.g. 'doi' or 'arxiv')
        print(result[0]['method'])              # Print the method used to find the identifier

    Parameters
    ----------
    target : string
        Relative or absolute path of a .pdf file or a directory containing pdf files

    Returns
    -------
    results, dictionary or list of dictionaries (or None if an error occured)
        The output is a single dictionary if target is a file, or a list of dictionaries if target is a directory,
        each element of the list describing one file. Each dictionary has the following keys

        result['identifier'] = DOI or other identifier (or None if nothing is found)
        result['identifier_type'] = string specifying the type of identifier (e.g. 'doi' or 'arxiv')
        result['validation_info'] = Additional info on the paper. If config.get('webvalidation') = True, then result['validation_info']
                                    will typically contain raw bibtex data for this paper. Otherwise it will just contain True
        result['path'] = path of the pdf file
        result['method'] = method used to find the identifier

    '''

    logger = logging.getLogger("pdf2doi")

    # Check if path is valid
    if not (path.exists(target)):
        logger.error(f"{target} is not a valid path to a file or a directory.")
        return None

    # Check if target is a directory
    # If yes, we look for all the .pdf files inside it, and for each of them
    # we call again this function
    if path.isdir(target):
        logger.info(f"Looking for pdf files in the folder {target}...")
        pdf_files = [f for f in listdir(target) if (f.lower()).endswith('.pdf')]
        numb_files = len(pdf_files)

        if numb_files == 0:
            logger.error("No pdf files found in this folder.")
            return None

        logger.info(f"Found {numb_files} pdf files.")
        if not (
        target.endswith(config.get('separator'))):  # Make sure the path ends with "\" or "/" (according to the OS)
            target = target + config.get('separator')

        identifiers_found = []  # For each pdf file in the target folder we will store a dictionary inside this list
        for f in pdf_files:
            logger.info("................")
            file = target + f
            # For each file we call again this function, but now the input argument target is set to the path of the file
            result = pdf2doi(target=file)
            logger.info(result['identifier'])
            identifiers_found.append(result)

        logger.info("................")

        return identifiers_found

    # If target is not a directory, we check that it is an existing file and that it ends with .pdf
    else:
        filename = target
        logger.info(f"Trying to retrieve a DOI/identifier for the file: {filename}")
        if not path.exists(filename):
            logger.error(f"'{filename}' is not a valid file.")
            return None
        if not (filename.lower()).endswith('.pdf'):
            logger.error("The file must have .pdf extension.")
            return None
        result = pdf2doi_singlefile(filename)
        if result['identifier'] == None:
            logger.error("It was not possible to find a valid identifier for this file.")

        if (config.get('save_identifier_metadata')) == True:
            if result['identifier'] and not (result['method'] == "document_infos"):
                finders.add_found_identifier_to_metadata(filename, result['identifier'])

        return result  # This will be a dictionary with all entries as None


def pdf2doi_singlefile(file):
    """
    Try to find an identifier of the file specified by the input argument file.  This function does not check wheter filename is a valid path to a pdf file.

    Parameters
    ----------
    file : ether a string or an object file
                if it's a string, it is the absolute path of a single .pdf file

    Returns
    -------
    result, dictionary
        The output is a single dictionary with the following keys

        result['identifier'] = DOI or other identifier (or None if nothing is found)
        result['identifier_type'] = string specifying the type of identifier (e.g. 'doi' or 'arxiv')
        result['validation_info'] = Additional info on the paper. If config.get('webvalidation') = True, then result['validation_info']
                                    will typically contain raw bibtex data for this paper. Otherwise it will just contain True
        result['path'] = path of the pdf file
        result['method'] = method used to find the identifier

    """

    logger = logging.getLogger("pdf2doi")

    result = {'identifier': None}

    try:
        with open(file, 'rb') as f:
            result = __find_doi(f)
    except TypeError:
        try:
            result = __find_doi(file)
        except Exception:
            logger.exception("File processing error")
    except Exception:
        logger.exception("File(open) processing error")

    return result


def __find_doi(file: io.IOBase) -> dict:
    logger = logging.getLogger("pdf2doi")

    # Several methods are now applied to find a valid identifier in the .pdf file identified by filename

    # First method: we look into the pdf metadata (in the current implementation this is done
    # via the getDocumentInfo() method of the library PyPdf) and see if any of them is a string which containts a
    # valid identifier inside it. We first look for the elements of the dictionary with keys '/doi' or /pdf2doi_identifier'(if the they exist),
    # and then any other field of the dictionary
    logger.info(f"Method #1: Looking for a valid identifier in the document infos...")
    result = finders.find_identifier(file, method="document_infos", keysToCheckFirst=['/doi', '/pdf2doi_identifier'])
    if result['identifier']:
        return result

    # Second method: We look for a DOI or arxiv ID inside the filename
    logger.info(f"Method #2: Looking for a valid identifier in the file name...")
    result = finders.find_identifier(file, method="filename")
    if result['identifier']:
        return result

    # Third method: We look in the plain text of the pdf and try to find something that matches a valid identifier.
    logger.info(f"Method #3: Looking for a valid identifier in the document text...")
    result = finders.find_identifier(file, method="document_text")
    if result['identifier']:
        return result

    # Fourth method: We look for possible titles of the paper, do a google search with them,
    # open the first results and look for identifiers in the plain text of the searcg results.
    logger.info(f"Method #4: Looking for possible publication titles...")
    result = finders.find_identifier(file, method="title_google")
    if result['identifier']:
        return result

    # Fifth method: We extract the first N characters from the file (where N is set by config.get('N_characters_in_pdf')) and we use it as
    # a query for a google seaerch. We open the first results and look for identifiers in the plain text of the searcg results.
    logger.info(
        f"Method #5: Trying to do a google search with the first {config.get('N_characters_in_pdf')} characters of this pdf file...")
    result = finders.find_identifier(file, method="first_N_characters_google")
    if result['identifier']:
        return result

    #If execution arrived to this point, it means that no identifier was found. We still return the dictionary returned by the last attempt, for further processing
    #In this case result['identifier']=None
    return result 


def save_identifiers(filename_identifiers, results, clipboard=False):
    ''' Write all identifiers contained in the input list 'results' into a text file with a path specified by filename_identifiers (if filename_identifiers is a
        valid string) and/or into the clipboard (if clipboard = True).

    Parameters
    ----------
    filename_identifiers : string
        Absolute path of the target file. If equal to '' or False, nothing is stored on file.
    results : list of dictionaries
        Each element of the list 'results' describes a .pdf file, and contains the pdf identifier and other infos.
    clipboard : boolean, optional
        If set to True, the identifiers are stored in the clipboard. Default is False.

    Returns
    -------
    None.
    '''
    logger = logging.getLogger("pdf2doi")

    # If a string was passed via the args.filename_identifiers, we create the full path of the file where identifiers will be saved
    if isinstance(filename_identifiers, str):
        path_filename_identifiers = path.dirname(results[0]['path']) + config.get('separator') + filename_identifiers
        try:
            text = ''
            for result in results:
                if result['validation_info']:
                    text = text + '{:<15s} {:<40s} {:<10s}\n'.format(result['identifier_type'], result['identifier'],
                                                                     result['path'])
                else:
                    text = text + '{:<15s} {:<40s} {:<10s}\n'.format('n.a.', 'n.a.', result['path'])
            with open(path_filename_identifiers, "w", encoding="utf-8") as text_file:
                text_file.write(text)
            logger.info(f'All found identifiers were saved in the file {filename_identifiers}')
        except Exception as e:
            logger.error(e)
            logger.error(f'A problem occurred when trying to write into the file {filename_identifiers}')

    # If clipboard is set to true, we copy all identifiers into the clipboard
    if clipboard:
        import pyperclip
        try:
            text = ''
            for result in results:
                if result['validation_info']:
                    text = text + result['identifier'] + '\n'
            pyperclip.copy(text)
            logger.info(f'All found identifiers have been stored in the system clipboard')
        except Exception as e:
            logger.error(e)
            logger.error(f'A problem occurred when trying to write into the system clipboard')


def main():
    parser = argparse.ArgumentParser(
        description="Retrieves the DOI or other identifiers (e.g. arXiv) from pdf files of a publications.",
        epilog="")

    parser.add_argument("path",
                        help="Relative path of the target pdf file or of the targe folder.",
                        metavar="path",
                        nargs='*')
    parser.add_argument("-v",
                        "--verbose",
                        help="Increase verbosity. By default (i.e. when not using -v), only a table with the found identifiers will be printed as output.",
                        action="store_true")
    parser.add_argument("-nws",
                        "--no_web_search",
                        help="Disable any method to find identifiers which requires internet searches (e.g. queries to google).",
                        action="store_true")
    parser.add_argument("-nwv",
                        "--no_web_validation",
                        help="Disable the online validation of identifiers (e.g., via queries to http://dx.doi.org/).",
                        action="store_true")
    parser.add_argument("-nostore",
                        "--no_store_identifier_metadata",
                        help="By default, anytime an identifier is found it is added to the metadata of the pdf file (if not present yet). By using this additional option, the identifier is not stored in the file metadata.",
                        action="store_true")
    parser.add_argument("-no_arxiv2doi",
                        help="If a valid arXiv ID is found for a given .pdf file, by default pdf2doi will try to also look for a DOI (either because the paper has been published in a journal or because arXiv has assigned to it a DOI of the form \"10.48550/arXivID\")." +
                            " By adding this command, the arXiv ID is instead always returned.",
                        action="store_true")
    parser.add_argument('-id',
                        help=f"Stores the string IDENTIFIER in the metadata of the target pdf file, with key \'/pdf2doi_identifier\'. Note: when this argument is passed, all other arguments (except for the path to the pdf file)" +
                             " are ignored. ",
                        action="store", dest="identifier", type=str, default=False)
    parser.add_argument("-id_input_box",  # When called with this argument, an input box is generated in order
                        # to acquire a string from the user, which is then stored in the metadata
                        # of the target pdf file
                        # This is normally used when calling pdf2doi by right-clicking on a .pdf file in Windows
                        help=argparse.SUPPRESS,
                        action="store_true")
    parser.add_argument('-google',
                        help=f"Set how many results should be considered when doing a google search for the DOI (default={str(config.get('numb_results_google_search'))}).",
                        action="store", dest="google_results", type=int)
    parser.add_argument("-s",
                        "--save_identifiers_file",
                        dest="filename_identifiers",
                        help="Save all the identifiers found in the target folder in a text file inside the same folder with name specified by FILENAME_IDENTIFIERS. This option is only available when a folder is targeted.",
                        action="store")
    parser.add_argument("-clip",
                        "--save_doi_clipboard",
                        action="store_true",
                        help="Store all found DOI/identifiers into the clipboard.")

    parser.add_argument("-install--right--click",
                        dest="install_right_click",
                        action="store_true",
                        help="Add a shortcut to pdf2doi in the right-click context menu of Windows. You can copy the identifier and/or bibtex entry of a pdf file (or all pdf files in a folder) into the clipboard by just right clicking on it! NOTE: this feature is only available on Windows.")
    parser.add_argument("-uninstall--right--click",
                        dest="uninstall_right_click",
                        action="store_true",
                        help="Uninstall the right-click context menu functionalities. NOTE: this feature is only available on Windows.")

    args = parser.parse_args()

    # Setup logging
    config.set('verbose',
               args.verbose)  # store the desired verbose level in the global config of pdf2doi. This will automatically update the logger verbosity

    # If the command -install--right--click was specified, it sets the right keys in the system registry
    if args.install_right_click:
        config.set('verbose', True)
        import pdf2doi.utils_registry as utils_registry
        utils_registry.install_right_click()
        return
    if args.uninstall_right_click:
        config.set('verbose', True)
        import pdf2doi.utils_registry as utils_registry
        utils_registry.uninstall_right_click()
        return
    if isinstance(args.path, list):
        if len(args.path) > 0:
            target = args.path[0]
        else:
            target = ""
    else:
        target = args.path

    if target == "":
        print("Error: the following arguments are required: path. Type \'pdf2doi --h\' for a list of commands.")
        return

    if not (path.exists(target)):
        print(f"Error: {target} is not a valid path to a file or a directory.")
        return None

    # If the command -id_input_box was specified, we generate an input box, ask for a string from the user and
    # save it in args.identifier. In this way, the next if block is triggered
    if args.id_input_box:
        import easygui
        identifier = easygui.enterbox(
            f"Please specify the identifier (i.e. DOI or arxiv ID) of the file:\n '{target}' \n\n(this will be stored in the file metadata labelled '\identifier')")
        args.identifier = identifier
    # If the command -id was specified, and if the user provided a valid string, we call the sub-routine to store the string passed by the user into the metadata of the file indicated by the user
    # Nothing else will be done
    if args.identifier or args.identifier == "":
        if isinstance(args.identifier, str):
            result = finders.add_found_identifier_to_metadata(target, args.identifier)
            if args.id_input_box and len(result) > 1:
                if result[0] == False:
                    easygui.msgbox(result[1])
        return

    config.set('replace_arxivID_by_DOI_when_available', not(args.no_arxiv2doi))
    config.set('websearch', not (args.no_web_search))
    config.set('webvalidation', not (args.no_web_validation))
    config.set('save_identifier_metadata', not (args.no_store_identifier_metadata))

    if args.google_results:
        config.set('numb_results_google_search', args.google_results)
    results = pdf2doi(target=target)

    if not results:
        return
    if not isinstance(results, list):
        results = [results]
    for result in results:
        if result['identifier']:
            print('{:<15s} {:<40s} {:<10s}\n'.format(result['identifier_type'], result['identifier'], result['path']))
        else:
            print('{:<15s} {:<40s} {:<10s}\n'.format('n.a.', 'n.a.', result['path']))

            # Call the function save_identifiers. If args.filename_identifiers is a valid string, it will save all found identifiers in a text file with that name.
    # If args.save_doi_clipboard is true, it will copy all identifiers into the clipboard. Otherwise, it will do nothing
    save_identifiers(args.filename_identifiers, results, args.save_doi_clipboard)

    return


if __name__ == '__main__':
    main()


import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'),encoding ='unicode_escape') as f:
    long_description = f.read()

with open("requirements.txt") as f:
    required_packages = f.read().splitlines()

setuptools.setup(name='pdf2doi',
      version='1.5rc7',
      description = 'A  python library/command-line tool to extract the DOI or other identifiers of a scientific paper from a pdf file.',
      long_description = long_description,
      long_description_content_type = 'text/markdown',
      url = 'https://github.com/MicheleCotrufo/pdf2doi',
      author = 'Michele Cotrufo',
      author_email = 'michele.cotrufo@gmail.com',
      license = 'MIT',
      entry_points = {
        'console_scripts': ["pdf2doi = pdf2doi.main:main"],
      },
      packages=['pdf2doi'],
      install_requires = required_packages,
      zip_safe = False)